from typing import Dict, Optional, List
from datetime import date
from dateutil.relativedelta import relativedelta
from sipametrics.internal.constants import COUNTRY_ISO_TO_ID_MAP, TIME_TO_MATURITY_BUCKETS, CURRENCY_MAP
from sipametrics.internal.enums import Product, AppName, TiccsProfiles
from sipametrics.endpoints import TAXONOMIES_URL


async def metrics_translation(entity_id: str, metric_id: str):
    params = {"metrics": [{"indexIds": [entity_id], "typeIds": [metric_id]}]}
    return params


async def infra_equity_comparable_translation(
    metric: str,
    currency: Optional[str] = None,
    age_in_months: Optional[int] = None,
    end_date: Optional[date] = None,
    window_in_years: Optional[int] = None,
    industrial_activities: Optional[List[str]] = None,
    business_risk: Optional[str] = None,
    corporate_structure: Optional[str] = None,
    countries: Optional[List[str]] = None,
    size: Optional[str] = None,
    leverage: Optional[str] = None,
    profitability: Optional[str] = None,
    investment: Optional[str] = None,
    time_to_maturity: Optional[str] = None,
    type: str = "mean",
):
    start_date_str = _calc_start_date(end_date=end_date, window_in_years=window_in_years)
    end_date_str = end_date.strftime("%Y-%m-%d") if end_date else None

    params = {
        "metric": metric,
        "startDate": start_date_str,
        "endDate": end_date_str,
        "age": age_in_months,
    }

    ticcs_profile_map = _create_ticcs_profile_map(
        industrial_activities=industrial_activities,
        business_risk=business_risk,
        corporate_structure=corporate_structure,
    )
    params["ticcsProfiles"] = ticcs_profile_map

    factors_profile_map = _create_infra_factors_profile_map(
        currency=currency,
        countries=countries,
        size=size,
        leverage=leverage,
        profitability=profitability,
        investment=investment,
        time_to_maturity=time_to_maturity,
        type=type,
    )
    params["factorsProfiles"] = factors_profile_map
    return params


async def infra_debt_comparable_translation(
    metric: str,
    currency: Optional[str] = None,
    age_in_months: Optional[int] = None,
    end_date: Optional[date] = None,
    window_in_years: Optional[int] = None,
    industrial_activities: Optional[List[str]] = None,
    business_risk: Optional[str] = None,
    corporate_structure: Optional[str] = None,
    countries: Optional[List[str]] = None,
    face_value: Optional[str] = None,
    debt_time_to_maturity: Optional[str] = None,
    type: str = "mean",
) -> Dict:
    start_date_str = _calc_start_date(end_date=end_date, window_in_years=window_in_years)
    end_date_str = end_date.strftime("%Y-%m-%d") if end_date else None

    ticcs_profiles = _create_ticcs_profile_map(
        industrial_activities=industrial_activities,
        business_risk=business_risk,
        corporate_structure=corporate_structure,
    )

    factors_profiles = _create_infra_factors_profile_map(
        countries=countries,
        face_value=face_value,
        debt_time_to_maturity=debt_time_to_maturity,
        type=type,
    )

    return {
        "metric": metric,
        "startDate": start_date_str,
        "endDate": end_date_str,
        "age": age_in_months,
        "ticcsProfiles": ticcs_profiles,
        "factorsProfiles": factors_profiles,
    }


async def term_structure_translation(country: str, date: str, maturity_date: str):
    params = {"countryName": country, "valueDate": date, "maturityDate": maturity_date}
    return params


async def private_equity_comparable_translation(
    metric: str,
    currency: Optional[str] = None,
    age_in_months: Optional[int] = None,
    end_date: Optional[date] = None,
    window_in_years: Optional[int] = None,
    industrial_activities: Optional[List[str]] = None,
    revenue_models: Optional[List[str]] = None,
    customer_models: Optional[List[str]] = None,
    lifecycle_phases: Optional[List[str]] = None,
    value_chain_types: Optional[List[str]] = None,
    countries: Optional[List[str]] = None,
    size: Optional[str] = None,
    growth: Optional[str] = None,
    leverage: Optional[str] = None,
    profits: Optional[str] = None,
    country_risk: Optional[List[str]] = None,
    universe: Optional[str] = None,
    factor_weight: Optional[str] = None,
    type: Optional[str] = "mean",
    intersect_peccs: Optional[bool] = None,
) -> Dict:
    start_date_str = _calc_start_date(end_date=end_date, window_in_years=window_in_years)
    end_date_str = end_date.strftime("%Y-%m-%d") if end_date else None

    peccs_country_profile = {
        "industrialActivities": industrial_activities,
        "revenueModel": revenue_models,
        "customerModel": customer_models,
        "lifecyclePhases": lifecycle_phases,
        "valueChainTypes": value_chain_types,
        "countries": countries,
    }
    peccs_country_profile = {key: value for key, value in peccs_country_profile.items() if value is not None}

    factors_profiles = _create_equity_factors_profile_map(
        size=size,
        growth=growth,
        leverage=leverage,
        profits=profits,
        country_risk=country_risk,
    )

    params = {
        "metric": metric,
        "currency": CURRENCY_MAP.get(currency or "USD", 840),
        "startDate": start_date_str,
        "endDate": end_date_str,
        "age": age_in_months,
        "peccsCountryProfile": peccs_country_profile,
        "factorsProfiles": factors_profiles,
        "universe": universe,
        "factorWeight": factor_weight,
        "operation": type,
        "intersectPeccs": intersect_peccs,
    }

    params = {key: value for key, value in params.items() if value is not None}
    return params


async def private_equity_comparable_boundaries_translation(
    metric: str,
    age_in_months: Optional[int] = None,
    end_date: Optional[date] = None,
    window_in_years: Optional[int] = None,
    industrial_activities: Optional[List[str]] = None,
    revenue_models: Optional[List[str]] = None,
    customer_models: Optional[List[str]] = None,
    lifecycle_phases: Optional[List[str]] = None,
    value_chain_types: Optional[List[str]] = None,
    countries: Optional[List[str]] = None,
    universe: Optional[str] = None,
    factor_weight: Optional[float] = None,
    factor_name: Optional[str] = None,
) -> Dict:

    start_date_str = _calc_start_date(end_date=end_date, window_in_years=window_in_years)
    end_date_str = end_date.strftime("%Y-%m-%d") if end_date else None

    peccs_country_profile = {
        "industrialActivities": industrial_activities,
        "revenueModel": revenue_models,
        "customerModel": customer_models,
        "lifecyclePhases": lifecycle_phases,
        "valueChainTypes": value_chain_types,
        "countries": countries,
    }
    peccs_country_profile = {key: value for key, value in peccs_country_profile.items() if value is not None}

    params = {
        "metric": metric,
        "ageInMonths": age_in_months,
        "startDate": start_date_str,
        "endDate": end_date_str,
        "universe": universe,
        "factorWeight": factor_weight,
        "factorName": factor_name,
        "peccsCountryProfile": peccs_country_profile,
    }
    params = {key: value for key, value in params.items() if value is not None}

    return params


async def taxonomies_translation(taxonomy: str, pillar: str):
    url = TAXONOMIES_URL.format(taxonomy=taxonomy)
    query_params = {"pillar": pillar}
    return url, query_params


async def indices_catalogue_translation(app, product):
    app_filter_map = {
        AppName.INDICES.value: {"marketIndices": 1} if product == Product.SIPA_METRICS.value else {"indexApp": 1},
        AppName.VALUATION.value: {"analytics": 1} if product == Product.SIPA_METRICS.value else {"assetValuation": 1},
        AppName.CLIMATE.value: {"indexApp": 1, "assetValuation": 1},
    }
    return app_filter_map.get(app, {})


def _create_ticcs_profile_map(
    industrial_activities: Optional[List[str]] = None,
    business_risk: Optional[str] = None,
    corporate_structure: Optional[str] = None,
) -> List:
    ticcs_profile_map = {}
    if industrial_activities:
        ticcs_profile_map[TiccsProfiles.INDUSTRIAL_SUPERCLASS.value] = industrial_activities
    if business_risk:
        ticcs_profile_map[TiccsProfiles.BUSINESS_MODEL.value] = business_risk
    if corporate_structure:
        ticcs_profile_map[TiccsProfiles.CORPORATE_STRUCTURE.value] = corporate_structure

    return [{"profile": {key: value}} for key, value in ticcs_profile_map.items()]


def _create_equity_factors_profile_map(
    size: Optional[str] = None,
    growth: Optional[str] = None,
    leverage: Optional[str] = None,
    profits: Optional[str] = None,
    country_risk: Optional[List[str]] = None,
) -> List:
    factors_profiles = []

    if size:
        if size[0] in ["Q", "T"]:
            factors_profiles.append(
                {
                    "factorName": "Size",
                    "quintile": _convert_to_quintile_or_time_series(size),
                }
            )
        else:
            factors_profiles.append(
                {
                    "factorName": "Size",
                    "Size": size,
                }
            )
    if growth:
        if growth[0] in ["Q", "T"]:
            factors_profiles.append(
                {
                    "factorName": "Growth",
                    "quintile": _convert_to_quintile_or_time_series(growth),
                }
            )
        else:
            factors_profiles.append(
                {
                    "factorName": "Growth",
                    "growth": growth,
                }
            )
    if leverage:
        if leverage[0] in ["Q", "T"]:
            factors_profiles.append(
                {
                    "factorName": "Leverage",
                    "quintile": _convert_to_quintile_or_time_series(leverage),
                }
            )
        else:
            factors_profiles.append(
                {
                    "factorName": "Leverage",
                    "leverage": leverage,
                }
            )
    if profits:
        if profits[0] in ["Q", "T"]:
            factors_profiles.append(
                {
                    "factorName": "Profits",
                    "quintile": _convert_to_quintile_or_time_series(profits),
                }
            )
        else:
            factors_profiles.append(
                {
                    "factorName": "Profits",
                    "profits": profits,
                }
            )
    if country_risk:
        factors_profiles.append(
            {
                "factorName": "TermSpread",
                "countries": country_risk,
            }
        )

    return factors_profiles


def _create_infra_factors_profile_map(
    currency: Optional[str] = None,
    countries: Optional[List[str]] = None,
    size: Optional[str] = None,
    leverage: Optional[str] = None,
    profitability: Optional[str] = None,
    investment: Optional[str] = None,
    time_to_maturity: Optional[str] = None,
    debt_time_to_maturity: Optional[str] = None,
    face_value: Optional[str] = None,
    type: Optional[str] = None,
) -> List:
    factors_profiles = []
    if countries:
        converted_countries = [COUNTRY_ISO_TO_ID_MAP.get(country, "Unknown") for country in countries]
        factors_profiles.append(
            {
                "profile": {
                    "factor": "countries",
                    "countries": converted_countries,
                }
            }
        )
    if size:
        if size[0] in ["Q", "T"]:
            factors_profiles.append(
                {
                    "profile": {
                        "factor": "size",
                        "quintile": _convert_to_quintile_or_time_series(size),
                        "currency": CURRENCY_MAP.get(currency or "USD", 840),
                    }
                }
            )
        else:
            factors_profiles.append(
                {
                    "profile": {
                        "factor": "size",
                        "size": size,
                        "currency": CURRENCY_MAP.get(currency or "USD", 840),
                    }
                }
            )

    if leverage:
        if leverage[0] in ["Q", "T"]:
            factors_profiles.append(
                {
                    "profile": {
                        "factor": "leverage",
                        "quintile": _convert_to_quintile_or_time_series(leverage),
                    }
                }
            )
        else:
            factors_profiles.append(
                {
                    "profile": {
                        "factor": "leverage",
                        "leverage": leverage,
                    }
                }
            )

    if profitability:
        if profitability[0] in ["Q", "T"]:
            factors_profiles.append(
                {
                    "profile": {
                        "factor": "profitability",
                        "quintile": _convert_to_quintile_or_time_series(profitability),
                    }
                }
            )
        else:
            factors_profiles.append(
                {
                    "profile": {
                        "factor": "profitability",
                        "profitability": profitability,
                    }
                }
            )

    if investment:
        if investment[0] in ["Q", "T"]:
            factors_profiles.append(
                {
                    "profile": {
                        "factor": "investment",
                        "quintile": _convert_to_quintile_or_time_series(investment),
                    }
                }
            )
        else:
            factors_profiles.append(
                {
                    "profile": {
                        "factor": "investment",
                        "investment": investment,
                    }
                }
            )

    if time_to_maturity:
        if time_to_maturity[0] in ["Q", "T"]:
            factors_profiles.append(
                {
                    "profile": {
                        "factor": "timeToMaturity",
                        "timeToMaturityBucket": _convert_to_quintile_or_time_series(time_to_maturity),
                    }
                }
            )
        else:
            factors_profiles.append(
                {
                    "profile": {
                        "factor": "timeToMaturity",
                        "timeToMaturity": time_to_maturity,
                    }
                }
            )

    if debt_time_to_maturity:
        if debt_time_to_maturity[0] in ["Q", "T"]:
            factors_profiles.append(
                {
                    "profile": {
                        "factor": "debtTimeToMaturity",
                        "timeToMaturityBucket": _convert_to_quintile_or_time_series(debt_time_to_maturity),
                    }
                }
            )
        else:
            factors_profiles.append(
                {
                    "profile": {
                        "factor": "debtTimeToMaturity",
                        "timeToMaturity": debt_time_to_maturity,
                    }
                }
            )

    if face_value:
        if face_value[0] in ["Q", "T"]:
            factors_profiles.append(
                {
                    "profile": {
                        "factor": "faceValue",
                        "quintile": _convert_to_quintile_or_time_series(face_value),
                    }
                }
            )
        else:
            factors_profiles.append(
                {
                    "profile": {
                        "factor": "faceValue",
                        "faceValue": face_value,
                    }
                }
            )

    return factors_profiles


def _convert_to_quintile_or_time_series(value: str):
    if value.upper().startswith("Q"):
        return int(value[1:])
    elif value.upper().startswith("T"):
        return TIME_TO_MATURITY_BUCKETS.get(value, value)
    else:
        return float(value) if value.isdigit() else value


def _calc_start_date(end_date: Optional[date], window_in_years: Optional[int]) -> Optional[str]:
    if end_date and window_in_years:
        start_date = end_date - relativedelta(years=window_in_years)
        return start_date.strftime("%Y-%m-%d")
    return None



from datetime import date
from typing import Dict, Optional, List
from sipametrics.base_client import BaseClient
from sipametrics.internal.translators import (
    indices_catalogue_translation,
    metrics_translation,
    infra_equity_comparable_translation,
    infra_debt_comparable_translation,
    private_equity_comparable_boundaries_translation,
    term_structure_translation,
    taxonomies_translation,
    private_equity_comparable_translation,
)
from sipametrics.endpoints import (
    METRICS_URL,
    INFRA_EQUITY_COMPARABLE_URL,
    INFRA_DEBT_COMPARABLE_URL,
    TERMSTRUCTURE_URL,
    PRIVATE_EQUITY_COMPARABLE_URL,
    PRIVATE_EQUITY_COMPARABLE_BOUNDARIES_URL,
    INDICES_CATALOGUE_URL,
    COUNTRIES_URL,
    PRIVATE_EQUITY_REGION_TREE_URL,
)


class SipaMetricsService:
    def __init__(self, api_key: str, api_secret: str):
        self.client = BaseClient(api_key, api_secret)

    async def close(self):
        await self.client.close()

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_value, tb):
        await self.client.close()

    async def metrics(self, entity_id: str, metric_id: str):
        """Returns the requested metrics."""
        params = await metrics_translation(entity_id, metric_id)
        return await self.client._post(METRICS_URL, data=params)

    async def infra_equity_comparable(
        self,
        metric: str,
        currency: Optional[str] = None,
        age_in_months: Optional[int] = None,
        end_date: Optional[date] = None,
        window_in_years: Optional[int] = None,
        industrial_activities: Optional[List[str]] = None,
        business_risk: Optional[str] = None,
        corporate_structure: Optional[str] = None,
        countries: Optional[List[str]] = None,
        size: Optional[str] = None,
        leverage: Optional[str] = None,
        profitability: Optional[str] = None,
        investment: Optional[str] = None,
        time_to_maturity: Optional[str] = None,
        type: str = "mean",
    ) -> Dict:
        """
        Perform a comparable computation for private infra equity. This involves finding datapoints which have similar TICCS classifications and factor values (the comparables dataset) and averaging the metric values.
        """
        params = await infra_equity_comparable_translation(
            metric=metric,
            currency=currency,
            age_in_months=age_in_months,
            end_date=end_date,
            window_in_years=window_in_years,
            industrial_activities=industrial_activities,
            business_risk=business_risk,
            corporate_structure=corporate_structure,
            countries=countries,
            size=size,
            leverage=leverage,
            profitability=profitability,
            investment=investment,
            time_to_maturity=time_to_maturity,
            type=type,
        )
        return await self.client._post(INFRA_EQUITY_COMPARABLE_URL, data=params)

    async def infra_debt_comparable(
        self,
        metric: str,
        currency: Optional[str] = None,
        age_in_months: Optional[int] = None,
        end_date: Optional[date] = None,
        window_in_years: Optional[int] = None,
        industrial_activities: Optional[List[str]] = None,
        business_risk: Optional[str] = None,
        corporate_structure: Optional[str] = None,
        countries: Optional[List[str]] = None,
        face_value: Optional[str] = None,
        time_to_maturity: Optional[str] = None,
        type: str = "mean",
    ) -> Dict:
        """
        Perform a comparable computation for private infra debt. This involves finding datapoints which have similar TICCS classifications and factor values (the comparables dataset) and averaging the metric values
        """
        params = await infra_debt_comparable_translation(
            metric=metric,
            currency=currency,
            age_in_months=age_in_months,
            end_date=end_date,
            window_in_years=window_in_years,
            industrial_activities=industrial_activities,
            business_risk=business_risk,
            corporate_structure=corporate_structure,
            countries=countries,
            face_value=face_value,
            debt_time_to_maturity=time_to_maturity,
            type=type,
        )
        return await self.client._post(INFRA_DEBT_COMPARABLE_URL, data=params)

    async def term_structure(self, country: str, date: str, maturity_date: str) -> Dict:
        """Query annualised risk-free rate for a given country and maturity date on the curve."""
        params = await term_structure_translation(
            country=country,
            date=date,
            maturity_date=maturity_date,
        )
        return await self.client._post(TERMSTRUCTURE_URL, data=params)

    async def private_equity_comparable(
        self,
        metric: str,
        currency: Optional[str] = None,
        age_in_months: Optional[int] = None,
        end_date: Optional[date] = None,
        window_in_years: Optional[int] = None,
        industrial_activities: Optional[List[str]] = None,
        revenue_models: Optional[List[str]] = None,
        customer_models: Optional[List[str]] = None,
        lifecycle_phases: Optional[List[str]] = None,
        value_chain_types: Optional[List[str]] = None,
        countries: Optional[List[str]] = None,
        size: Optional[str] = None,
        growth: Optional[str] = None,
        leverage: Optional[str] = None,
        profits: Optional[str] = None,
        country_risk: Optional[List[str]] = None,
        universe: Optional[str] = None,
        factor_weight: Optional[str] = None,
        type: Optional[str] = "mean",
        intersect_peccs: Optional[bool] = None,
    ) -> Dict:
        """Perform a comparable computation for private equity. This involves finding datapoints which have similar PECCS classifications and factor values (the comparables dataset) and averaging the metric values."""
        params = await private_equity_comparable_translation(
            metric=metric,
            currency=currency,
            age_in_months=age_in_months,
            end_date=end_date,
            window_in_years=window_in_years,
            industrial_activities=industrial_activities,
            revenue_models=revenue_models,
            customer_models=customer_models,
            lifecycle_phases=lifecycle_phases,
            value_chain_types=value_chain_types,
            countries=countries,
            size=size,
            growth=growth,
            leverage=leverage,
            profits=profits,
            country_risk=country_risk,
            universe=universe,
            factor_weight=factor_weight,
            type=type,
            intersect_peccs=intersect_peccs,
        )
        return await self.client._post(PRIVATE_EQUITY_COMPARABLE_URL, data=params)

    async def private_equity_comparable_boundaries(
        self,
        metric: str,
        age_in_months: Optional[int] = None,
        end_date: Optional[date] = None,
        window_in_years: Optional[int] = None,
        industrial_activities: Optional[List[str]] = None,
        revenue_models: Optional[List[str]] = None,
        customer_models: Optional[List[str]] = None,
        lifecycle_phases: Optional[List[str]] = None,
        value_chain_types: Optional[List[str]] = None,
        countries: Optional[List[str]] = None,
        universe: Optional[str] = None,
        factor_weight: Optional[float] = None,
        factor_name: Optional[str] = None,
    ) -> Dict:
        """
        Handles the request to compute private equity comparable boundaries.
        """
        # Translate parameters using the translation function
        params = await private_equity_comparable_boundaries_translation(
            metric=metric,
            age_in_months=age_in_months,
            end_date=end_date,
            window_in_years=window_in_years,
            industrial_activities=industrial_activities,
            revenue_models=revenue_models,
            customer_models=customer_models,
            lifecycle_phases=lifecycle_phases,
            value_chain_types=value_chain_types,
            countries=countries,
            universe=universe,
            factor_weight=factor_weight,
            factor_name=factor_name,
        )
        return await self.client._post(PRIVATE_EQUITY_COMPARABLE_BOUNDARIES_URL, data=params)

    async def indices_catalogue(self, product: Optional[str] = None, app: Optional[str] = None) -> Dict:
        """Return all available indices catalogue."""
        query_params = {}
        if product:
            query_params["product"] = product
        additional_query = await indices_catalogue_translation(product=product, app=app)
        query_params.update(additional_query)
        return await self.client._get(INDICES_CATALOGUE_URL, params=query_params)

    async def taxonomies(self, taxonomy: str, pillar: str) -> Dict:
        """Return requested taxonomies mapping."""
        allowed_list = ["ticcs", "peccs", "ticcs+", "ticcs-eu", "ticcs-nace"]
        if taxonomy not in allowed_list:
            raise ValueError(f"Invalid taxonomy: {taxonomy}. Allowed values are: {allowed_list}")
        url, query_params = await taxonomies_translation(taxonomy=taxonomy, pillar=pillar)

        return await self.client._get(url, params=query_params)

    async def countries(self) -> Dict:
        """Return list of countries."""
        return await self.client._get(COUNTRIES_URL)

    async def private_equity_region_tree(self) -> Dict:
        """Return hierarchical view of regions and countries."""
        return await self.client._get(PRIVATE_EQUITY_REGION_TREE_URL)



PUBLIC_API_URL = "https://indices-backend.edhecinfra.com"
METRICS_URL = f"{PUBLIC_API_URL}/data-service/v2/queries/metrics"
INFRA_EQUITY_COMPARABLE_URL = f"{PUBLIC_API_URL}/compute-service/v2/pi-comparables/average/by/profiles"
INFRA_DEBT_COMPARABLE_URL = f"{PUBLIC_API_URL}/compute-service/v2/pi-comparables/average/by/profiles"
TERMSTRUCTURE_URL = f"{PUBLIC_API_URL}/data-service/v2/yields"
INDICES_CATALOGUE_URL = f"{PUBLIC_API_URL}/data-service/v2/lists/indices"
TAXONOMIES_URL = f"{PUBLIC_API_URL}/data-service/v2/lists/taxonomy/{{taxonomy}}"
COUNTRIES_URL = f"{PUBLIC_API_URL}/data-service/v2/yields/countries"
PRIVATE_EQUITY_COMPARABLE_URL = f"{PUBLIC_API_URL}/compute-service/v2/pe-comparables/average/by/profiles"
PRIVATE_EQUITY_COMPARABLE_BOUNDARIES_URL = f"{PUBLIC_API_URL}/compute-service/v2/pe-comparables/boundaries/by/profiles"
PRIVATE_EQUITY_REGION_TREE_URL = f"{PUBLIC_API_URL}/compute-service/v2/pe-comparables/region-tree"
PUBLIC_API_VERSION = "v2"
SOURCE = "sipaMetrics-python-sdk"



from enum import Enum


class TiccsProfiles(Enum):
    INDUSTRIAL_SUPERCLASS = "industrialSuperclass"
    BUSINESS_MODEL = "businessModel"
    CORPORATE_STRUCTURE = "corporateStructure"
    COUNTRY = "countries"


class Product(Enum):
    SIPA_METRICS = "pe"
    INFRA_METRICS = "pi"


class AppName(Enum):
    INDICES = "indices"
    VALUATION = "valuation"
    CLIMATE = "climate"
